/**
 * Import file contents as arrays of objects
 * 
 * ## Installation
 *
 *    > npm install nxus-data-loader --save
 *
 * ## Options
 *
 * All responses accept an options argument, in addition to any parser-specific options
 * you can indicate:
 *  * `mapping`: object of {field: newField} name mappings
 *  * `identityFields`: for importing to models, array of fields to be used for createOrUpdate query
 *  * `truncate`: for importing to models, true/false for deleting existing collection data before import. Ignored if `identityFields` is provided.
 *  * `strict`: defaults to true. Only import columns/data that matches the attribute names for the model. Set to false to import everything.
 * 
 * ## Events
 *
 * You can modify the records during import with the following specific events:
 *  * `records.type`: e.g. `app.get('data-loader').after('records.csv', (results) => {})`
 *  * `record.type`: e.g. `app.get('data-loader).after('record.csv', (one) => {})`
 *  * `models.identity`: e.g. `app.get('data-loader').after('models.user', (results) => {})`
 *  * `model.identity`: e.g. `app.get('data-loader).after('model.user', (user) => {})`
 *
 * record* events occur after parsing and name mapping
 * model* events occur after record events and before models are created/updated.
 * 
 * # API
 * -----
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _underscore = require('underscore');

var _underscore2 = _interopRequireDefault(_underscore);

var _multer = require('multer');

var _multer2 = _interopRequireDefault(_multer);

var _morph = require('morph');

var _morph2 = _interopRequireDefault(_morph);

var _CSVParser = require('./CSVParser');

var _CSVParser2 = _interopRequireDefault(_CSVParser);

var _CSVExporter = require('./CSVExporter');

var _CSVExporter2 = _interopRequireDefault(_CSVExporter);

var _ArcJSONParser = require('./ArcJSONParser');

var _ArcJSONParser2 = _interopRequireDefault(_ArcJSONParser);

var _GeoJSONParser = require('./GeoJSONParser');

var _GeoJSONParser2 = _interopRequireDefault(_GeoJSONParser);

var _GeoJSONExporter = require('./GeoJSONExporter');

var _GeoJSONExporter2 = _interopRequireDefault(_GeoJSONExporter);

var _JSONParser = require('./JSONParser');

var _JSONParser2 = _interopRequireDefault(_JSONParser);

var _JSONExporter = require('./JSONExporter');

var _JSONExporter2 = _interopRequireDefault(_JSONExporter);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

_bluebird2.default.promisifyAll(_fs2.default);

var _defaultImportOptions = {
  truncate: true,
  identityFields: [],
  strict: true
};

var DataLoader = function () {
  function DataLoader(app) {
    _classCallCheck(this, DataLoader);

    this.app = app;
    this._parsers = {};
    this._exporters = {};
    this.storage = app.get('storage');
    app.get('data-loader').use(this).gather('parser').gather('exporter').gather('uploadPath').respond('export').respond('import').respond('importFile').respond('importToModel').respond('importFileToModel');

    new _CSVParser2.default(app);
    new _CSVExporter2.default(app);
    new _ArcJSONParser2.default(app);
    new _GeoJSONParser2.default(app);
    new _GeoJSONExporter2.default(app);
    new _JSONParser2.default(app);
    new _JSONExporter2.default(app);
  }

  /**
   * Provide a parser for a particular type (file extension)
   * @param {string} type The type (e.g. 'html') this renderer should handle
   * @param {function} handler Function to receive (content, options) and return parsed array of result objects
   */


  _createClass(DataLoader, [{
    key: 'parser',
    value: function parser(type, handler) {
      this.on('records.' + type, function (args) {
        return args;
      });
      this.on('record.' + type, function (args) {
        return args;
      });
      this._parsers[type] = handler;
    }
    /**
     * Provide an exporter for a particular type (file extension)
     * @param {string} type The type (e.g. 'html') this exporter creates
     * @param {function} handler Function to receive (content, options) and return formatted output content
     */

  }, {
    key: 'exporter',
    value: function exporter(type, handler) {
      this._exporters[type] = handler;
    }
  }, {
    key: 'uploadPath',
    value: function uploadPath(path, field) {
      var dest = arguments.length <= 2 || arguments[2] === undefined ? process.cwd() + '/uploads/' : arguments[2];

      var upload = (0, _multer2.default)({ dest: dest });
      this.app.get('router').middleware(path, upload.single(field));
    }

    /**
     * Request formattted output based on type
     * @param {string} type The type (e.g. 'html') of the output content
     * @param {[object]} records The records to export
     * @param {object} opts Options for the exporter context
     * @return {Promise} String of formatted output
     */

  }, {
    key: 'export',
    value: function _export(type, records, opts) {
      if (opts === undefined) opts = {};
      if (!this._exporters[type]) throw new Error('No matching exporter found: ' + type);
      return _bluebird2.default.resolve(this._exporters[type](records, opts));
    }

    /**
     * Request parsed results based on type
     * @param {string} type The type (e.g. 'html') of the content
     * @param {string} content The contents to parse 
     * @param {object} opts Options for the parser context
     * @return {Promise} Array of parsed result objects
     */

  }, {
    key: 'import',
    value: function _import(type, content, opts) {
      var _this = this;

      if (opts === undefined) opts = {};
      if (!this._parsers[type]) throw new Error('No matching parser found: ' + type);
      return _bluebird2.default.resolve(this._parsers[type](content, opts)).then(function (results) {
        var records = _bluebird2.default.mapSeries(_this._mappingNames(results, opts), function (record) {
          return _this.emit('record.' + type, record);
        });
        return _this.emit('records.' + type, records);
      });
    }

    /**
     * Request parsed results from a file path
     * @param {string} filename The filename to read and parse
     * @param {object} opts Options for the parser context
     * @return {Promise} Array of parsed result objects
     */

  }, {
    key: 'importFile',
    value: function importFile(filename, opts) {
      var _this2 = this;

      return _fs2.default.readFileAsync(filename).then(function (content) {
        content = content.toString();
        var type = _path2.default.extname(filename).replace(".", "");
        if (opts && opts.type) {
          type = opts.type;
        }
        return _this2.import(type, content, opts);
      });
    }

    /**
     * Import string contents to a model
     * @param {string} model The identity of the model to populate
     * @param {string} type The type (e.g. 'html') of the content
     * @param {string} content The contents to parse 
     * @param {object} opts Options for the parser context
     * @return {Promise} Array of instances
     */

  }, {
    key: 'importToModel',
    value: function importToModel(model, type, content, opts) {
      var _this3 = this;

      return this.request('import', type, content, opts).then(function (results) {
        return _this3._modelImport(model, results, opts);
      });
    }

    /**
     * Import file contents to a model
     * @param {string} model The identity of the model to populate
     * @param {string} filename The filename to read and parse
     * @param {object} opts Options for the parser context
     * @return {Promise} Array of instances
     */

  }, {
    key: 'importFileToModel',
    value: function importFileToModel(model, filename, opts) {
      var _this4 = this;

      return this.request('importFile', filename, opts).then(function (results) {
        return _this4._modelImport(model, results, opts);
      });
    }

    // Internal

  }, {
    key: '_modelImport',
    value: function _modelImport(model_id, results, opts) {
      var _this5 = this;

      var loader = this.app.get('data-loader');
      //stub default handlers on "model" and "models" events
      if (loader.listenerCount('models.' + model_id) == 0) {
        this.on('models.' + model_id, function (args) {
          return args;
        });
      }
      if (loader.listenerCount('model.' + model_id) == 0) {
        this.on('model.' + model_id, function (args) {
          return args;
        });
      }
      if (opts === undefined) opts = _defaultImportOptions;

      if (typeof opts.strict != 'undefined' && opts.strict === false) {
        results = _underscore2.default.map(results, function (r) {
          var values = {};
          _underscore2.default.each(r, function (value, key) {
            values[_morph2.default.toSnake(key)] = value;
          });
          return values;
        });
      }

      return _bluebird2.default.mapSeries(results, function (record) {
        return _this5.emit('model.' + model_id, record);
      }).then(function (records) {
        return _this5.emit('models.' + model_id, records).then(function (results) {
          _this5.app.log.debug("about to call the storage for model " + model_id);
          return _this5.storage.getModel(model_id).then(function (model) {
            var model_attrs = _underscore2.default.keys(model._attributes);
            var identityFields = opts.identityFields;
            var hasIdentity = !_underscore2.default.isEmpty(identityFields);
            var importResults = function importResults() {
              return _bluebird2.default.mapSeries(results, function (r) {
                var action = null;
                if (typeof opts.strict != 'undefined' && opts.strict === false) {
                  var values = r;
                } else {
                  var values = _underscore2.default.pick.apply(_underscore2.default, [r].concat(_toConsumableArray(model_attrs)));
                }
                if (hasIdentity) {
                  var criteria = _underscore2.default.pick.apply(_underscore2.default, [values].concat(_toConsumableArray(identityFields)));
                  action = _this5._storeResultsWithModel(model, values, criteria);
                } else {
                  action = _this5._storeResultsWithModel(model, values);
                }
                return action;
              });
            };

            if (opts.truncate && !hasIdentity) {
              return model.destroy({}).then(importResults);
            } else {
              return importResults();
            }
          }).catch(function (e) {
            _this5.app.log.error("Error imporing " + model_id, e, e.details);
          });
        });
      });
    }

    /**
     * Simple do-storage function
     * to help mock up storage for tests, other applications.
     * @param  {[type]} model          [description]
     * @param  {[type]} values         [description]
     * @param  {[type]} uniqueCriteria [description]
     * @return {[type]}                [description]
     */

  }, {
    key: '_storeResultsWithModel',
    value: function _storeResultsWithModel(model, values, uniqueCriteria) {
      if (uniqueCriteria) return model.createOrUpdate(uniqueCriteria, values);else return model.create(values);
    }
  }, {
    key: '_mappingNames',
    value: function _mappingNames(results, options) {
      var mapping = options.mapping;
      if (mapping === undefined || _underscore2.default.isEmpty(mapping)) {
        return results;
      }
      var oldNames = _underscore2.default.keys(mapping);
      return _underscore2.default.map(results, function (r) {
        _underscore2.default.mapObject(mapping, function (k, v) {
          r[k] = r[v];
        });
        return _underscore2.default.omit.apply(_underscore2.default, [r].concat(_toConsumableArray(oldNames)));
      });
    }
  }]);

  return DataLoader;
}();

exports.default = DataLoader;